"use strict";
Object.defineProperty(exports, Symbol.toStringTag, { value: "Module" });
const path = require("node:path");
const fs = require("node:fs");
const zod = require("zod");
const config = require("./filesystem/virtual/config.cjs");
const configSchema = zod.z.object({
  target: zod.z.enum(["react", "solid"]).optional().default("react"),
  virtualRouteConfig: config.virtualRootRouteSchema.or(zod.z.string()).optional(),
  routeFilePrefix: zod.z.string().optional(),
  routeFileIgnorePrefix: zod.z.string().optional().default("-"),
  routeFileIgnorePattern: zod.z.string().optional(),
  routesDirectory: zod.z.string().optional().default("./src/routes"),
  generatedRouteTree: zod.z.string().optional().default("./src/routeTree.gen.ts"),
  quoteStyle: zod.z.enum(["single", "double"]).optional().default("single"),
  semicolons: zod.z.boolean().optional().default(false),
  disableTypes: zod.z.boolean().optional().default(false),
  addExtensions: zod.z.boolean().optional().default(false),
  disableLogging: zod.z.boolean().optional().default(false),
  disableManifestGeneration: zod.z.boolean().optional().default(false),
  enableRouteTreeFormatting: zod.z.boolean().optional().default(true),
  __enableAPIRoutesGeneration: zod.z.boolean().optional(),
  // Internal flag to be turned on for TanStack Start
  apiBase: zod.z.string().optional().default("/api"),
  routeTreeFileHeader: zod.z.array(zod.z.string()).optional().default([
    "/* eslint-disable */",
    "// @ts-nocheck",
    "// noinspection JSUnusedGlobalSymbols"
  ]),
  routeTreeFileFooter: zod.z.array(zod.z.string()).optional().default([]),
  autoCodeSplitting: zod.z.boolean().optional(),
  indexToken: zod.z.string().optional().default("index"),
  routeToken: zod.z.string().optional().default("route"),
  pathParamsAllowedCharacters: zod.z.array(zod.z.enum([";", ":", "@", "&", "=", "+", "$", ","])).optional(),
  customScaffolding: zod.z.object({
    routeTemplate: zod.z.string().optional(),
    lazyRouteTemplate: zod.z.string().optional(),
    apiTemplate: zod.z.string().optional()
  }).optional(),
  experimental: zod.z.object({
    // TODO: This has been made stable and is now "autoCodeSplitting". Remove in next major version.
    enableCodeSplitting: zod.z.boolean().optional()
  }).optional()
});
function resolveConfigPath({ configDirectory }) {
  return path.resolve(configDirectory, "tsr.config.json");
}
function getConfig(inlineConfig = {}, configDirectory) {
  if (configDirectory === void 0) {
    configDirectory = process.cwd();
  }
  const configFilePathJson = resolveConfigPath({ configDirectory });
  const exists = fs.existsSync(configFilePathJson);
  let config2;
  if (exists) {
    config2 = configSchema.parse({
      ...JSON.parse(fs.readFileSync(configFilePathJson, "utf-8")),
      ...inlineConfig
    });
  } else {
    config2 = configSchema.parse(inlineConfig);
  }
  if (config2.disableTypes) {
    config2.generatedRouteTree = config2.generatedRouteTree.replace(
      /\.(ts|tsx)$/,
      ".js"
    );
  }
  if (configDirectory) {
    if (path.isAbsolute(configDirectory)) {
      config2.routesDirectory = path.resolve(
        configDirectory,
        config2.routesDirectory
      );
      config2.generatedRouteTree = path.resolve(
        configDirectory,
        config2.generatedRouteTree
      );
    } else {
      config2.routesDirectory = path.resolve(
        process.cwd(),
        configDirectory,
        config2.routesDirectory
      );
      config2.generatedRouteTree = path.resolve(
        process.cwd(),
        configDirectory,
        config2.generatedRouteTree
      );
    }
  }
  validateConfig(config2);
  return config2;
}
function validateConfig(config2) {
  var _a;
  if (typeof ((_a = config2.experimental) == null ? void 0 : _a.enableCodeSplitting) !== "undefined") {
    const message = `
------
⚠️ ⚠️ ⚠️
ERROR: The "experimental.enableCodeSplitting" flag has been made stable and is now "autoCodeSplitting". Please update your configuration file to use "autoCodeSplitting" instead of "experimental.enableCodeSplitting".
------
`;
    console.error(message);
    throw new Error(message);
  }
  if (config2.indexToken === config2.routeToken) {
    throw new Error(
      `The "indexToken" and "routeToken" options must be different.`
    );
  }
  if (config2.routeFileIgnorePrefix && config2.routeFileIgnorePrefix.trim() === "_") {
    throw new Error(
      `The "routeFileIgnorePrefix" cannot be an underscore ("_"). This is a reserved character used to denote a pathless route. Please use a different prefix.`
    );
  }
  return config2;
}
exports.configSchema = configSchema;
exports.getConfig = getConfig;
exports.resolveConfigPath = resolveConfigPath;
//# sourceMappingURL=config.cjs.map
