"use strict";
Object.defineProperty(exports, Symbol.toStringTag, { value: "Module" });
const path = require("node:path");
const fs = require("node:fs");
const fsp = require("node:fs/promises");
const utils = require("./utils.cjs");
const getRouteNodes$1 = require("./filesystem/physical/getRouteNodes.cjs");
const getRouteNodes = require("./filesystem/virtual/getRouteNodes.cjs");
const rootPathId = require("./filesystem/physical/rootPathId.cjs");
const template = require("./template.cjs");
function _interopNamespaceDefault(e) {
  const n = Object.create(null, { [Symbol.toStringTag]: { value: "Module" } });
  if (e) {
    for (const k in e) {
      if (k !== "default") {
        const d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: () => e[k]
        });
      }
    }
  }
  n.default = e;
  return Object.freeze(n);
}
const fs__namespace = /* @__PURE__ */ _interopNamespaceDefault(fs);
const fsp__namespace = /* @__PURE__ */ _interopNamespaceDefault(fsp);
const CONSTANTS = {
  // When changing this, you'll want to update the import in `start-api-routes/src/index.ts#defaultAPIFileRouteHandler`
  APIRouteExportVariable: "APIRoute"
};
let latestTask = 0;
const routeGroupPatternRegex = /\(.+\)/g;
const possiblyNestedRouteGroupPatternRegex = /\([^/]+\)\/?/g;
let isFirst = false;
let skipMessage = false;
async function generator(config, root) {
  const ROUTE_TEMPLATE = template.getTargetTemplate(config.target);
  const logger = utils.logging({ disabled: config.disableLogging });
  logger.log("");
  if (!isFirst) {
    logger.log("♻️  Generating routes...");
    isFirst = true;
  } else if (skipMessage) {
    skipMessage = false;
  } else {
    logger.log("♻️  Regenerating routes...");
  }
  const taskId = latestTask + 1;
  latestTask = taskId;
  const checkLatest = () => {
    if (latestTask !== taskId) {
      skipMessage = true;
      return false;
    }
    return true;
  };
  const start = Date.now();
  const TYPES_DISABLED = config.disableTypes;
  const ENABLED_API_ROUTES_GENERATION = config.__enableAPIRoutesGeneration ?? false;
  let getRouteNodesResult;
  if (config.virtualRouteConfig) {
    getRouteNodesResult = await getRouteNodes.getRouteNodes(config, root);
  } else {
    getRouteNodesResult = await getRouteNodes$1.getRouteNodes(config, root);
  }
  const { rootRouteNode, routeNodes: beforeRouteNodes } = getRouteNodesResult;
  if (rootRouteNode === void 0) {
    let errorMessage = `rootRouteNode must not be undefined. Make sure you've added your root route into the route-tree.`;
    if (!config.virtualRouteConfig) {
      errorMessage += `
Make sure that you add a "${rootPathId.rootPathId}.${config.disableTypes ? "js" : "tsx"}" file to your routes directory.
Add the file in: "${config.routesDirectory}/${rootPathId.rootPathId}.${config.disableTypes ? "js" : "tsx"}"`;
    }
    throw new Error(errorMessage);
  }
  const preRouteNodes = utils.multiSortBy(beforeRouteNodes, [
    (d) => d.routePath === "/" ? -1 : 1,
    (d) => {
      var _a;
      return (_a = d.routePath) == null ? void 0 : _a.split("/").length;
    },
    (d) => d.filePath.match(new RegExp(`[./]${config.indexToken}[.]`)) ? 1 : -1,
    (d) => d.filePath.match(
      /[./](component|errorComponent|pendingComponent|loader|lazy)[.]/
    ) ? 1 : -1,
    (d) => d.filePath.match(new RegExp(`[./]${config.routeToken}[.]`)) ? -1 : 1,
    (d) => {
      var _a;
      return ((_a = d.routePath) == null ? void 0 : _a.endsWith("/")) ? -1 : 1;
    },
    (d) => d.routePath
  ]).filter((d) => ![`/${rootPathId.rootPathId}`].includes(d.routePath || ""));
  const routeTree = [];
  const routePiecesByPath = {};
  const onlyAPIRouteNodes = preRouteNodes.filter((d) => {
    if (!ENABLED_API_ROUTES_GENERATION) {
      return false;
    }
    if (d._fsRouteType !== "api") {
      return false;
    }
    return true;
  });
  const onlyGeneratorRouteNodes = preRouteNodes.filter((d) => {
    if (ENABLED_API_ROUTES_GENERATION) {
      if (d._fsRouteType === "api") {
        return false;
      }
    }
    return true;
  });
  const routeNodes = [];
  const handleRootNode = async (node) => {
    if (!node) {
      return;
    }
    const routeCode = fs__namespace.readFileSync(node.fullPath, "utf-8");
    if (!routeCode) {
      const _rootTemplate = ROUTE_TEMPLATE.rootRoute;
      const replaced = await template.fillTemplate(config, _rootTemplate.template(), {
        tsrImports: _rootTemplate.imports.tsrImports(),
        tsrPath: rootPathId.rootPathId,
        tsrExportStart: _rootTemplate.imports.tsrExportStart(),
        tsrExportEnd: _rootTemplate.imports.tsrExportEnd()
      });
      await utils.writeIfDifferent(
        node.fullPath,
        "",
        // Empty string because the file doesn't exist yet
        replaced,
        {
          beforeWrite: () => {
            logger.log(`🟡 Creating ${node.fullPath}`);
          }
        }
      );
    }
  };
  await handleRootNode(rootRouteNode);
  const handleNode = async (node) => {
    var _a, _b, _c, _d, _e;
    utils.resetRegex(routeGroupPatternRegex);
    let parentRoute = hasParentRoute(routeNodes, node, node.routePath);
    if ((parentRoute == null ? void 0 : parentRoute.isVirtualParentRoute) && ((_a = parentRoute.children) == null ? void 0 : _a.length)) {
      const possibleParentRoute = hasParentRoute(
        parentRoute.children,
        node,
        node.routePath
      );
      if (possibleParentRoute) {
        parentRoute = possibleParentRoute;
      }
    }
    if (parentRoute) node.parent = parentRoute;
    node.path = determineNodePath(node);
    const trimmedPath = utils.trimPathLeft(node.path ?? "");
    const split = trimmedPath.split("/");
    const lastRouteSegment = split[split.length - 1] ?? trimmedPath;
    node.isNonPath = lastRouteSegment.startsWith("_") || routeGroupPatternRegex.test(lastRouteSegment);
    node.cleanedPath = removeGroups(
      utils.removeUnderscores(removeLayoutSegments(node.path)) ?? ""
    );
    if (!node.isVirtualParentRoute && !node.isVirtual) {
      const routeCode = fs__namespace.readFileSync(node.fullPath, "utf-8");
      const escapedRoutePath = ((_b = node.routePath) == null ? void 0 : _b.replaceAll("$", "$$")) ?? "";
      let replaced = routeCode;
      const tRouteTemplate = ROUTE_TEMPLATE.route;
      const tLazyRouteTemplate = ROUTE_TEMPLATE.lazyRoute;
      if (!routeCode) {
        if (node._fsRouteType === "lazy") {
          replaced = await template.fillTemplate(
            config,
            (((_c = config.customScaffolding) == null ? void 0 : _c.lazyRouteTemplate) || ((_d = config.customScaffolding) == null ? void 0 : _d.routeTemplate)) ?? tLazyRouteTemplate.template(),
            {
              tsrImports: tLazyRouteTemplate.imports.tsrImports(),
              tsrPath: escapedRoutePath,
              tsrExportStart: tLazyRouteTemplate.imports.tsrExportStart(escapedRoutePath),
              tsrExportEnd: tLazyRouteTemplate.imports.tsrExportEnd()
            }
          );
        } else if (
          // Creating a new normal route file
          ["layout", "static"].some(
            (d) => d === node._fsRouteType
          ) || [
            "component",
            "pendingComponent",
            "errorComponent",
            "loader"
          ].every((d) => d !== node._fsRouteType)
        ) {
          replaced = await template.fillTemplate(
            config,
            ((_e = config.customScaffolding) == null ? void 0 : _e.routeTemplate) ?? tRouteTemplate.template(),
            {
              tsrImports: tRouteTemplate.imports.tsrImports(),
              tsrPath: escapedRoutePath,
              tsrExportStart: tRouteTemplate.imports.tsrExportStart(escapedRoutePath),
              tsrExportEnd: tRouteTemplate.imports.tsrExportEnd()
            }
          );
        }
      } else {
        replaced = routeCode.replace(
          /(FileRoute\(\s*['"])([^\s]*)(['"],?\s*\))/g,
          (_, p1, __, p3) => `${p1}${escapedRoutePath}${p3}`
        ).replace(
          new RegExp(
            `(import\\s*\\{.*)(create(Lazy)?FileRoute)(.*\\}\\s*from\\s*['"]@tanstack\\/${ROUTE_TEMPLATE.subPkg}['"])`,
            "gs"
          ),
          (_, p1, __, ___, p4) => `${p1}${node._fsRouteType === "lazy" ? "createLazyFileRoute" : "createFileRoute"}${p4}`
        ).replace(
          /create(Lazy)?FileRoute(\(\s*['"])([^\s]*)(['"],?\s*\))/g,
          (_, __, p2, ___, p4) => `${node._fsRouteType === "lazy" ? "createLazyFileRoute" : "createFileRoute"}${p2}${escapedRoutePath}${p4}`
        );
      }
      await utils.writeIfDifferent(node.fullPath, routeCode, replaced, {
        beforeWrite: () => {
          logger.log(`🟡 Updating ${node.fullPath}`);
        }
      });
    }
    if (!node.isVirtual && [
      "lazy",
      "loader",
      "component",
      "pendingComponent",
      "errorComponent"
    ].some((d) => d === node._fsRouteType)) {
      routePiecesByPath[node.routePath] = routePiecesByPath[node.routePath] || {};
      routePiecesByPath[node.routePath][node._fsRouteType === "lazy" ? "lazy" : node._fsRouteType === "loader" ? "loader" : node._fsRouteType === "errorComponent" ? "errorComponent" : node._fsRouteType === "pendingComponent" ? "pendingComponent" : "component"] = node;
      const anchorRoute = routeNodes.find((d) => d.routePath === node.routePath);
      if (!anchorRoute) {
        await handleNode({
          ...node,
          isVirtual: true,
          _fsRouteType: "static"
        });
      }
      return;
    }
    const cleanedPathIsEmpty = (node.cleanedPath || "").length === 0;
    const nonPathRoute = node._fsRouteType === "pathless_layout" && node.isNonPath;
    node.isVirtualParentRequired = node._fsRouteType === "pathless_layout" || nonPathRoute ? !cleanedPathIsEmpty : false;
    if (!node.isVirtual && node.isVirtualParentRequired) {
      const parentRoutePath = removeLastSegmentFromPath(node.routePath) || "/";
      const parentVariableName = utils.routePathToVariable(parentRoutePath);
      const anchorRoute = routeNodes.find(
        (d) => d.routePath === parentRoutePath
      );
      if (!anchorRoute) {
        const parentNode = {
          ...node,
          path: removeLastSegmentFromPath(node.path) || "/",
          filePath: removeLastSegmentFromPath(node.filePath) || "/",
          fullPath: removeLastSegmentFromPath(node.fullPath) || "/",
          routePath: parentRoutePath,
          variableName: parentVariableName,
          isVirtual: true,
          _fsRouteType: "layout",
          // layout since this route will wrap other routes
          isVirtualParentRoute: true,
          isVirtualParentRequired: false
        };
        parentNode.children = parentNode.children ?? [];
        parentNode.children.push(node);
        node.parent = parentNode;
        if (node._fsRouteType === "pathless_layout") {
          node.path = determineNodePath(node);
        }
        await handleNode(parentNode);
      } else {
        anchorRoute.children = anchorRoute.children ?? [];
        anchorRoute.children.push(node);
        node.parent = anchorRoute;
      }
    }
    if (node.parent) {
      if (!node.isVirtualParentRequired) {
        node.parent.children = node.parent.children ?? [];
        node.parent.children.push(node);
      }
    } else {
      routeTree.push(node);
    }
    routeNodes.push(node);
  };
  for (const node of onlyGeneratorRouteNodes) {
    await handleNode(node);
  }
  checkRouteFullPathUniqueness(
    preRouteNodes.filter(
      (d) => d.children === void 0 && ["api", "lazy"].every(
        (type) => type !== d._fsRouteType
      )
    ),
    config
  );
  const startAPIRouteNodes = checkStartAPIRoutes(
    onlyAPIRouteNodes,
    config
  );
  const handleAPINode = async (node) => {
    var _a, _b;
    const routeCode = fs__namespace.readFileSync(node.fullPath, "utf-8");
    const escapedRoutePath = ((_a = node.routePath) == null ? void 0 : _a.replaceAll("$", "$$")) ?? "";
    if (!routeCode) {
      const replaced = await template.fillTemplate(
        config,
        ((_b = config.customScaffolding) == null ? void 0 : _b.apiTemplate) ?? template.defaultAPIRouteTemplate,
        {
          tsrImports: "import { createAPIFileRoute } from '@tanstack/react-start/api';",
          tsrPath: escapedRoutePath,
          tsrExportStart: `export const ${CONSTANTS.APIRouteExportVariable} = createAPIFileRoute('${escapedRoutePath}')(`,
          tsrExportEnd: ");"
        }
      );
      await utils.writeIfDifferent(
        node.fullPath,
        "",
        // Empty string because the file doesn't exist yet
        replaced,
        {
          beforeWrite: () => {
            logger.log(`🟡 Creating ${node.fullPath}`);
          }
        }
      );
    } else {
      await utils.writeIfDifferent(
        node.fullPath,
        routeCode,
        routeCode.replace(
          /(createAPIFileRoute\(\s*['"])([^\s]*)(['"],?\s*\))/g,
          (_, p1, __, p3) => `${p1}${escapedRoutePath}${p3}`
        ),
        {
          beforeWrite: () => {
            logger.log(`🟡 Updating ${node.fullPath}`);
          }
        }
      );
    }
  };
  if (ENABLED_API_ROUTES_GENERATION) {
    for (const node of startAPIRouteNodes) {
      await handleAPINode(node);
    }
  }
  function buildRouteTreeConfig(nodes, depth = 1) {
    const children = nodes.map((node) => {
      var _a, _b;
      if (node._fsRouteType === "__root") {
        return;
      }
      if (node._fsRouteType === "pathless_layout" && !((_a = node.children) == null ? void 0 : _a.length)) {
        return;
      }
      const route = `${node.variableName}Route`;
      if ((_b = node.children) == null ? void 0 : _b.length) {
        const childConfigs = buildRouteTreeConfig(node.children, depth + 1);
        const childrenDeclaration = TYPES_DISABLED ? "" : `interface ${route}Children {
  ${node.children.map((child) => `${child.variableName}Route: typeof ${getResolvedRouteNodeVariableName(child)}`).join(",")}
}`;
        const children2 = `const ${route}Children${TYPES_DISABLED ? "" : `: ${route}Children`} = {
  ${node.children.map((child) => `${child.variableName}Route: ${getResolvedRouteNodeVariableName(child)}`).join(",")}
}`;
        const routeWithChildren = `const ${route}WithChildren = ${route}._addFileChildren(${route}Children)`;
        return [
          childConfigs,
          childrenDeclaration,
          children2,
          routeWithChildren
        ].join("\n\n");
      }
      return void 0;
    });
    return children.filter(Boolean).join("\n\n");
  }
  const routeConfigChildrenText = buildRouteTreeConfig(routeTree);
  const sortedRouteNodes = utils.multiSortBy(routeNodes, [
    (d) => {
      var _a;
      return ((_a = d.routePath) == null ? void 0 : _a.includes(`/${rootPathId.rootPathId}`)) ? -1 : 1;
    },
    (d) => {
      var _a;
      return (_a = d.routePath) == null ? void 0 : _a.split("/").length;
    },
    (d) => {
      var _a;
      return ((_a = d.routePath) == null ? void 0 : _a.endsWith(config.indexToken)) ? -1 : 1;
    },
    (d) => d
  ]);
  const imports = Object.entries({
    createFileRoute: sortedRouteNodes.some((d) => d.isVirtual),
    lazyFn: sortedRouteNodes.some(
      (node) => {
        var _a;
        return (_a = routePiecesByPath[node.routePath]) == null ? void 0 : _a.loader;
      }
    ),
    lazyRouteComponent: sortedRouteNodes.some(
      (node) => {
        var _a, _b, _c;
        return ((_a = routePiecesByPath[node.routePath]) == null ? void 0 : _a.component) || ((_b = routePiecesByPath[node.routePath]) == null ? void 0 : _b.errorComponent) || ((_c = routePiecesByPath[node.routePath]) == null ? void 0 : _c.pendingComponent);
      }
    )
  }).filter((d) => d[1]).map((d) => d[0]);
  const virtualRouteNodes = sortedRouteNodes.filter((d) => d.isVirtual);
  function getImportPath(node) {
    return utils.replaceBackslash(
      utils.removeExt(
        path.relative(
          path.dirname(config.generatedRouteTree),
          path.resolve(config.routesDirectory, node.filePath)
        ),
        config.addExtensions
      )
    );
  }
  const routeImports = [
    ...config.routeTreeFileHeader,
    `// This file was automatically generated by TanStack Router.
// You should NOT make any changes in this file as it will be overwritten.
// Additionally, you should also exclude this file from your linter and/or formatter to prevent it from being checked or modified.`,
    imports.length ? `import { ${imports.join(", ")} } from '${ROUTE_TEMPLATE.fullPkg}'
` : "",
    "// Import Routes",
    [
      `import { Route as rootRoute } from './${getImportPath(rootRouteNode)}'`,
      ...sortedRouteNodes.filter((d) => !d.isVirtual).map((node) => {
        return `import { Route as ${node.variableName}Import } from './${getImportPath(node)}'`;
      })
    ].join("\n"),
    virtualRouteNodes.length ? "// Create Virtual Routes" : "",
    virtualRouteNodes.map((node) => {
      return `const ${node.variableName}Import = createFileRoute('${node.routePath}')()`;
    }).join("\n"),
    "// Create/Update Routes",
    sortedRouteNodes.map((node) => {
      var _a, _b, _c, _d, _e, _f;
      const loaderNode = (_a = routePiecesByPath[node.routePath]) == null ? void 0 : _a.loader;
      const componentNode = (_b = routePiecesByPath[node.routePath]) == null ? void 0 : _b.component;
      const errorComponentNode = (_c = routePiecesByPath[node.routePath]) == null ? void 0 : _c.errorComponent;
      const pendingComponentNode = (_d = routePiecesByPath[node.routePath]) == null ? void 0 : _d.pendingComponent;
      const lazyComponentNode = (_e = routePiecesByPath[node.routePath]) == null ? void 0 : _e.lazy;
      return [
        `const ${node.variableName}Route = ${node.variableName}Import.update({
          ${[
          `id: '${node.path}'`,
          !node.isNonPath ? `path: '${node.cleanedPath}'` : void 0,
          `getParentRoute: () => ${((_f = node.parent) == null ? void 0 : _f.variableName) ?? "root"}Route`
        ].filter(Boolean).join(",")}
        }${TYPES_DISABLED ? "" : "as any"})`,
        loaderNode ? `.updateLoader({ loader: lazyFn(() => import('./${utils.replaceBackslash(
          utils.removeExt(
            path.relative(
              path.dirname(config.generatedRouteTree),
              path.resolve(config.routesDirectory, loaderNode.filePath)
            ),
            config.addExtensions
          )
        )}'), 'loader') })` : "",
        componentNode || errorComponentNode || pendingComponentNode ? `.update({
              ${[
          ["component", componentNode],
          ["errorComponent", errorComponentNode],
          ["pendingComponent", pendingComponentNode]
        ].filter((d) => d[1]).map((d) => {
          return `${d[0]}: lazyRouteComponent(() => import('./${utils.replaceBackslash(
            utils.removeExt(
              path.relative(
                path.dirname(config.generatedRouteTree),
                path.resolve(config.routesDirectory, d[1].filePath)
              ),
              config.addExtensions
            )
          )}'), '${d[0]}')`;
        }).join("\n,")}
            })` : "",
        lazyComponentNode ? `.lazy(() => import('./${utils.replaceBackslash(
          utils.removeExt(
            path.relative(
              path.dirname(config.generatedRouteTree),
              path.resolve(
                config.routesDirectory,
                lazyComponentNode.filePath
              )
            ),
            config.addExtensions
          )
        )}').then((d) => d.Route))` : ""
      ].join("");
    }).join("\n\n"),
    ...TYPES_DISABLED ? [] : [
      "// Populate the FileRoutesByPath interface",
      `declare module '${ROUTE_TEMPLATE.fullPkg}' {
  interface FileRoutesByPath {
    ${routeNodes.map((routeNode) => {
        var _a, _b;
        const filePathId = routeNode.routePath;
        return `'${filePathId}': {
          id: '${filePathId}'
          path: '${inferPath(routeNode)}'
          fullPath: '${inferFullPath(routeNode)}'
          preLoaderRoute: typeof ${routeNode.variableName}Import
          parentRoute: typeof ${routeNode.isVirtualParentRequired ? `${(_a = routeNode.parent) == null ? void 0 : _a.variableName}Route` : ((_b = routeNode.parent) == null ? void 0 : _b.variableName) ? `${routeNode.parent.variableName}Import` : "rootRoute"}
        }`;
      }).join("\n")}
  }
}`
    ],
    "// Create and export the route tree",
    routeConfigChildrenText,
    ...TYPES_DISABLED ? [] : [
      `export interface FileRoutesByFullPath {
  ${[...createRouteNodesByFullPath(routeNodes).entries()].map(
        ([fullPath, routeNode]) => {
          return `'${fullPath}': typeof ${getResolvedRouteNodeVariableName(routeNode)}`;
        }
      )}
}`,
      `export interface FileRoutesByTo {
  ${[...createRouteNodesByTo(routeNodes).entries()].map(([to, routeNode]) => {
        return `'${to}': typeof ${getResolvedRouteNodeVariableName(routeNode)}`;
      })}
}`,
      `export interface FileRoutesById {
  '__root__': typeof rootRoute,
  ${[...createRouteNodesById(routeNodes).entries()].map(([id, routeNode]) => {
        return `'${id}': typeof ${getResolvedRouteNodeVariableName(routeNode)}`;
      })}
}`,
      `export interface FileRouteTypes {
  fileRoutesByFullPath: FileRoutesByFullPath
  fullPaths: ${routeNodes.length > 0 ? [...createRouteNodesByFullPath(routeNodes).keys()].map((fullPath) => `'${fullPath}'`).join("|") : "never"}
  fileRoutesByTo: FileRoutesByTo
  to: ${routeNodes.length > 0 ? [...createRouteNodesByTo(routeNodes).keys()].map((to) => `'${to}'`).join("|") : "never"}
  id: ${[`'__root__'`, ...[...createRouteNodesById(routeNodes).keys()].map((id) => `'${id}'`)].join("|")}
  fileRoutesById: FileRoutesById
}`,
      `export interface RootRouteChildren {
  ${routeTree.map((child) => `${child.variableName}Route: typeof ${getResolvedRouteNodeVariableName(child)}`).join(",")}
}`
    ],
    `const rootRouteChildren${TYPES_DISABLED ? "" : ": RootRouteChildren"} = {
  ${routeTree.map((child) => `${child.variableName}Route: ${getResolvedRouteNodeVariableName(child)}`).join(",")}
}`,
    `export const routeTree = rootRoute._addFileChildren(rootRouteChildren)${TYPES_DISABLED ? "" : "._addFileTypes<FileRouteTypes>()"}`,
    ...config.routeTreeFileFooter
  ].filter(Boolean).join("\n\n");
  const createRouteManifest = () => {
    const routesManifest = {
      __root__: {
        filePath: rootRouteNode.filePath,
        children: routeTree.map((d) => d.routePath)
      },
      ...Object.fromEntries(
        routeNodes.map((d) => {
          var _a, _b;
          const filePathId = d.routePath;
          return [
            filePathId,
            {
              filePath: d.filePath,
              parent: ((_a = d.parent) == null ? void 0 : _a.routePath) ? d.parent.routePath : void 0,
              children: (_b = d.children) == null ? void 0 : _b.map((childRoute) => childRoute.routePath)
            }
          ];
        })
      )
    };
    return JSON.stringify(
      {
        routes: routesManifest
      },
      null,
      2
    );
  };
  const includeManifest = ["react", "solid"];
  const routeConfigFileContent = config.disableManifestGeneration || !includeManifest.includes(config.target) ? routeImports : [
    routeImports,
    "\n",
    "/* ROUTE_MANIFEST_START",
    createRouteManifest(),
    "ROUTE_MANIFEST_END */"
  ].join("\n");
  if (!checkLatest()) return;
  const existingRouteTreeContent = await fsp__namespace.readFile(path.resolve(config.generatedRouteTree), "utf-8").catch((err) => {
    if (err.code === "ENOENT") {
      return "";
    }
    throw err;
  });
  if (!checkLatest()) return;
  await fsp__namespace.mkdir(path.dirname(path.resolve(config.generatedRouteTree)), {
    recursive: true
  });
  if (!checkLatest()) return;
  const routeTreeWriteResult = await utils.writeIfDifferent(
    path.resolve(config.generatedRouteTree),
    config.enableRouteTreeFormatting ? await utils.format(existingRouteTreeContent, config) : existingRouteTreeContent,
    config.enableRouteTreeFormatting ? await utils.format(routeConfigFileContent, config) : routeConfigFileContent,
    {
      beforeWrite: () => {
        logger.log(`🟡 Updating ${config.generatedRouteTree}`);
      }
    }
  );
  if (routeTreeWriteResult && !checkLatest()) {
    return;
  }
  logger.log(
    `✅ Processed ${routeNodes.length === 1 ? "route" : "routes"} in ${Date.now() - start}ms`
  );
}
function removeGroups(s) {
  return s.replace(possiblyNestedRouteGroupPatternRegex, "");
}
function determineNodePath(node) {
  var _a;
  return node.path = node.parent ? ((_a = node.routePath) == null ? void 0 : _a.replace(node.parent.routePath ?? "", "")) || "/" : node.routePath;
}
function removeLastSegmentFromPath(routePath = "/") {
  const segments = routePath.split("/");
  segments.pop();
  return segments.join("/");
}
function removeLayoutSegments(routePath = "/") {
  const segments = routePath.split("/");
  const newSegments = segments.filter((segment) => !segment.startsWith("_"));
  return newSegments.join("/");
}
function hasParentRoute(routes, node, routePathToCheck) {
  if (!routePathToCheck || routePathToCheck === "/") {
    return null;
  }
  const sortedNodes = utils.multiSortBy(routes, [
    (d) => d.routePath.length * -1,
    (d) => d.variableName
  ]).filter((d) => d.routePath !== `/${rootPathId.rootPathId}`);
  for (const route of sortedNodes) {
    if (route.routePath === "/") continue;
    if (routePathToCheck.startsWith(`${route.routePath}/`) && route.routePath !== routePathToCheck) {
      return route;
    }
  }
  const segments = routePathToCheck.split("/");
  segments.pop();
  const parentRoutePath = segments.join("/");
  return hasParentRoute(routes, node, parentRoutePath);
}
const getResolvedRouteNodeVariableName = (routeNode) => {
  var _a;
  return ((_a = routeNode.children) == null ? void 0 : _a.length) ? `${routeNode.variableName}RouteWithChildren` : `${routeNode.variableName}Route`;
};
const createRouteNodesByFullPath = (routeNodes) => {
  return new Map(
    routeNodes.map((routeNode) => [inferFullPath(routeNode), routeNode])
  );
};
const createRouteNodesByTo = (routeNodes) => {
  return new Map(
    dedupeBranchesAndIndexRoutes(routeNodes).map((routeNode) => [
      inferTo(routeNode),
      routeNode
    ])
  );
};
const createRouteNodesById = (routeNodes) => {
  return new Map(
    routeNodes.map((routeNode) => {
      const id = routeNode.routePath ?? "";
      return [id, routeNode];
    })
  );
};
const inferFullPath = (routeNode) => {
  const fullPath = removeGroups(
    utils.removeUnderscores(removeLayoutSegments(routeNode.routePath)) ?? ""
  );
  return routeNode.cleanedPath === "/" ? fullPath : fullPath.replace(/\/$/, "");
};
const inferPath = (routeNode) => {
  var _a;
  return routeNode.cleanedPath === "/" ? routeNode.cleanedPath : ((_a = routeNode.cleanedPath) == null ? void 0 : _a.replace(/\/$/, "")) ?? "";
};
const inferTo = (routeNode) => {
  const fullPath = inferFullPath(routeNode);
  if (fullPath === "/") return fullPath;
  return fullPath.replace(/\/$/, "");
};
const dedupeBranchesAndIndexRoutes = (routes) => {
  return routes.filter((route) => {
    var _a;
    if ((_a = route.children) == null ? void 0 : _a.find((child) => child.cleanedPath === "/")) return false;
    return true;
  });
};
function checkUnique(routes, key) {
  const keys = routes.map((d) => d[key]);
  const uniqueKeys = new Set(keys);
  if (keys.length !== uniqueKeys.size) {
    const duplicateKeys = keys.filter((d, i) => keys.indexOf(d) !== i);
    const conflictingFiles = routes.filter(
      (d) => duplicateKeys.includes(d[key])
    );
    return conflictingFiles;
  }
  return void 0;
}
function checkRouteFullPathUniqueness(_routes, config) {
  const routes = _routes.map((d) => {
    const inferredFullPath = inferFullPath(d);
    return { ...d, inferredFullPath };
  });
  const conflictingFiles = checkUnique(routes, "inferredFullPath");
  if (conflictingFiles !== void 0) {
    const errorMessage = `Conflicting configuration paths were found for the following route${conflictingFiles.length > 1 ? "s" : ""}: ${conflictingFiles.map((p) => `"${p.inferredFullPath}"`).join(", ")}.
Please ensure each route has a unique full path.
Conflicting files: 
 ${conflictingFiles.map((d) => path.resolve(config.routesDirectory, d.filePath)).join("\n ")}
`;
    throw new Error(errorMessage);
  }
}
function checkStartAPIRoutes(_routes, config) {
  if (_routes.length === 0) {
    return [];
  }
  const routes = _routes.map((d) => {
    const routePath = utils.removeTrailingSlash(d.routePath ?? "");
    return { ...d, routePath };
  });
  const conflictingFiles = checkUnique(routes, "routePath");
  if (conflictingFiles !== void 0) {
    const errorMessage = `Conflicting configuration paths were found for the following API route${conflictingFiles.length > 1 ? "s" : ""}: ${conflictingFiles.map((p) => `"${p}"`).join(", ")}.
  Please ensure each API route has a unique route path.
Conflicting files: 
 ${conflictingFiles.map((d) => path.resolve(config.routesDirectory, d.filePath)).join("\n ")}
`;
    throw new Error(errorMessage);
  }
  return routes;
}
function startAPIRouteSegmentsFromTSRFilePath(src, config) {
  const routePath = utils.determineInitialRoutePath(src);
  const parts = routePath.replaceAll(".", "/").split("/").filter((p) => !!p && p !== config.indexToken);
  const segments = parts.map((part) => {
    if (part.startsWith("$")) {
      if (part === "$") {
        return { value: part, type: "splat" };
      }
      part.replaceAll("$", "");
      return { value: part, type: "param" };
    }
    return { value: part, type: "path" };
  });
  return segments;
}
exports.CONSTANTS = CONSTANTS;
exports.createRouteNodesByFullPath = createRouteNodesByFullPath;
exports.createRouteNodesById = createRouteNodesById;
exports.createRouteNodesByTo = createRouteNodesByTo;
exports.dedupeBranchesAndIndexRoutes = dedupeBranchesAndIndexRoutes;
exports.generator = generator;
exports.getResolvedRouteNodeVariableName = getResolvedRouteNodeVariableName;
exports.hasParentRoute = hasParentRoute;
exports.inferFullPath = inferFullPath;
exports.inferPath = inferPath;
exports.inferTo = inferTo;
exports.removeLastSegmentFromPath = removeLastSegmentFromPath;
exports.startAPIRouteSegmentsFromTSRFilePath = startAPIRouteSegmentsFromTSRFilePath;
//# sourceMappingURL=generator.cjs.map
