import * as fs from "node:fs";
import * as prettier from "prettier";
function multiSortBy(arr, accessors = [(d) => d]) {
  return arr.map((d, i) => [d, i]).sort(([a, ai], [b, bi]) => {
    for (const accessor of accessors) {
      const ao = accessor(a);
      const bo = accessor(b);
      if (typeof ao === "undefined") {
        if (typeof bo === "undefined") {
          continue;
        }
        return 1;
      }
      if (ao === bo) {
        continue;
      }
      return ao > bo ? 1 : -1;
    }
    return ai - bi;
  }).map(([d]) => d);
}
function cleanPath(path) {
  return path.replace(/\/{2,}/g, "/");
}
function trimPathLeft(path) {
  return path === "/" ? path : path.replace(/^\/{1,}/, "");
}
function logging(config) {
  function stripEmojis(str) {
    return str.replace(
      /[\p{Emoji_Presentation}\p{Extended_Pictographic}]/gu,
      ""
    );
  }
  function formatLogArgs(args) {
    if (process.env.CI) {
      return args.map(
        (arg) => typeof arg === "string" ? stripEmojis(arg) : arg
      );
    }
    return args;
  }
  return {
    log: (...args) => {
      if (!config.disabled) console.log(...formatLogArgs(args));
    },
    debug: (...args) => {
      if (!config.disabled) console.debug(...formatLogArgs(args));
    },
    info: (...args) => {
      if (!config.disabled) console.info(...formatLogArgs(args));
    },
    warn: (...args) => {
      if (!config.disabled) console.warn(...formatLogArgs(args));
    },
    error: (...args) => {
      if (!config.disabled) console.error(...formatLogArgs(args));
    }
  };
}
function removeLeadingSlash(path) {
  return path.replace(/^\//, "");
}
function removeTrailingSlash(s) {
  return s.replace(/\/$/, "");
}
function determineInitialRoutePath(routePath) {
  return cleanPath(`/${routePath.split(".").join("/")}`) || "";
}
function replaceBackslash(s) {
  return s.replaceAll(/\\/gi, "/");
}
function routePathToVariable(routePath) {
  var _a;
  return ((_a = removeUnderscores(routePath)) == null ? void 0 : _a.replace(/\/\$\//g, "/splat/").replace(/\$$/g, "splat").replace(/\$/g, "").split(/[/-]/g).map((d, i) => i > 0 ? capitalize(d) : d).join("").replace(/([^a-zA-Z0-9]|[.])/gm, "").replace(/^(\d)/g, "R$1")) ?? "";
}
function removeUnderscores(s) {
  return s == null ? void 0 : s.replaceAll(/(^_|_$)/gi, "").replaceAll(/(\/_|_\/)/gi, "/");
}
function capitalize(s) {
  if (typeof s !== "string") return "";
  return s.charAt(0).toUpperCase() + s.slice(1);
}
function removeExt(d, keepExtension = false) {
  return keepExtension ? d : d.substring(0, d.lastIndexOf(".")) || d;
}
async function writeIfDifferent(filepath, content, incomingContent, callbacks) {
  var _a, _b;
  if (content !== incomingContent) {
    (_a = callbacks == null ? void 0 : callbacks.beforeWrite) == null ? void 0 : _a.call(callbacks);
    fs.writeFileSync(filepath, incomingContent);
    (_b = callbacks == null ? void 0 : callbacks.afterWrite) == null ? void 0 : _b.call(callbacks);
    return true;
  }
  return false;
}
async function format(source, config) {
  const prettierOptions = {
    semi: config.semicolons,
    singleQuote: config.quoteStyle === "single",
    parser: "typescript"
  };
  return prettier.format(source, prettierOptions);
}
function resetRegex(regex) {
  regex.lastIndex = 0;
  return;
}
export {
  capitalize,
  cleanPath,
  determineInitialRoutePath,
  format,
  logging,
  multiSortBy,
  removeExt,
  removeLeadingSlash,
  removeTrailingSlash,
  removeUnderscores,
  replaceBackslash,
  resetRegex,
  routePathToVariable,
  trimPathLeft,
  writeIfDifferent
};
//# sourceMappingURL=utils.js.map
